from fastapi.testclient import TestClient

def run_embed_tests(client: TestClient):
    print('Running embedding tests...')
    response = client.post('/v1/embeddings', json={
        'model': 'svllm-embed',
        'input': ['hello', 'world'],
    })
    assert response.status_code == 200
    assert 'data' in response.json()
    assert isinstance(response.json()['data'], list)
    assert len(response.json()['data']) == 2
    assert 'embedding' in response.json()['data'][0]
    assert isinstance(response.json()['data'][0]['embedding'], list)
    assert len(response.json()['data'][0]['embedding']) > 0



from .__about__ import __version__
from .server import create_app
from .protocol import Chat, Complete, Embed, Model, Message, LengthException
from . import base, examples, openai

set_chat = base.set_chat
get_chat = base.get_chat

set_complete = base.set_complete
get_complete = base.get_complete

set_embed = base.set_embed
get_embed = base.get_embed

set_models = base.set_models
get_models = base.get_models

__all__ = [
    '__version__',
    'create_app',

    'set_chat',
    'get_chat',

    'set_complete',
    'get_complete',

    'set_embed',
    'get_embed',

    'set_models',
    'get_models',

    'base',
    'examples',

    'Chat',
    'Complete',
    'Embed',
    'Model',
    'Message',
    'LengthException',

    'openai',
]



from ..base import get_models as g_get_models
from .protocol import ModelList, ModelCard
from .helpers import create_500_error

def get_models() -> ModelList:
    try:
        data = list(map(lambda m: ModelCard(**m), g_get_models()))
        return ModelList(data=data)
    except Exception as e:
        return create_500_error(str(e))



import asyncio, inspect
from typing import List

from ..protocol import Embed
from ..base import get_embed as g_get_embed
from ..helpers import tokens
from . import helpers
from .protocol import Embedding, EmbeddingUsageInfo, EmbeddingsRequest, EmbeddingsResponse

def get_arguments(request: EmbeddingsRequest, sig: inspect.Signature) -> dict:
    arguments = {}
    if 'model' in sig.parameters:
        arguments['model'] = request.model
    if 'engine' in sig.parameters:
        arguments['engine'] = request.engine
    if 'user' in sig.parameters:
        arguments['user'] = request.user
    if 'encoding_format' in sig.parameters:
        arguments['encoding_format'] = request.encoding_format
    return arguments

async def embed_async(embed: Embed, kvargs: dict) -> List[List[float]]:
    result = await asyncio.to_thread(embed, **kvargs)
    return await result if inspect.iscoroutine(result) else result

async def create_embeddings(request: EmbeddingsRequest) -> EmbeddingsResponse:
    try:
        embed = g_get_embed()
        if not embed:
            return helpers.create_501_error('embed')

        embed_sig = inspect.signature(embed)
        arguments = get_arguments(request, embed_sig)

        input = request.input if isinstance(request.input, list) else [request.input]
        if 'input' in embed_sig.parameters:
            arguments['input'] = input

        embeddings: List[Embedding] = []
        for (i, data) in enumerate(await embed_async(embed, arguments)):
            embeddings.append(Embedding(embedding=data, index=i))

        prompt_tokens = sum(tokens(item) for item in input)

        return EmbeddingsResponse(
            model=request.model,
            data=embeddings,
            usage=EmbeddingUsageInfo(
                prompt_tokens=prompt_tokens,
                total_tokens=prompt_tokens,
            )
        )
    except Exception as e:
        return helpers.create_500_error(str(e))



from fastapi.responses import JSONResponse
from .protocol import ErrorResponse
from typing import Optional

def create_error(
        message: str,
        type: str,
        param: Optional[str] = None,
        code: Optional[str] = None,
        status: int = 500) -> JSONResponse:
    return JSONResponse(
        status_code=status,
        content=ErrorResponse(
            error={
                'message': message,
                'type': type,
                'param': param,
                'code': code
            }
        ).model_dump()
    )

def create_500_error(message: str) -> JSONResponse:
    return create_error(
        message=message,
        type='internal_error',
        code='500 INTERNAL_ERROR',
        status=500
    )

def create_501_error(func: str) -> JSONResponse:
    return create_error(
        message=f'{func} not implemented',
        type='not_implemented',
        code='501 NOT_IMPLEMENT',
        status=501
    )





import inspect
from fastapi.responses import StreamingResponse

from ..protocol import LengthException, Complete
from ..base import get_complete as g_get_complete
from ..helpers import agg, asyncgen, tokens
from . import helpers
from .protocol import (
    CompletionRequest,
    CompletionResponse,
    CompletionResponseChoice,
    CompletionResponseStreamChoice,
    CompletionStreamResponse,
    UsageInfo,
)

def get_arguments(request: CompletionRequest, sig: inspect.Signature) -> dict:
    arguments = {}
    if 'prompt' in sig.parameters:
        arguments['prompt'] = request.prompt
    if 'model' in sig.parameters:
        arguments['model'] = request.model
    if 'suffix' in sig.parameters:
        arguments['suffix'] = request.suffix
    if 'temperature' in sig.parameters:
        arguments['temperature'] = request.temperature
    if 'top_p' in sig.parameters:
        arguments['top_p'] = request.top_p
    if 'top_k' in sig.parameters:
        arguments['top_k'] = request.top_k
    if 'max_tokens' in sig.parameters:
        arguments['max_tokens'] = request.max_tokens
    if 'stop' in sig.parameters:
        arguments['stop'] = request.stop
    if 'stream' in sig.parameters:
        arguments['stream'] = request.stream
    if 'logprobs' in sig.parameters:
        arguments['logprobs'] = request.logprobs
    if 'echo' in sig.parameters:
        arguments['echo'] = request.echo
    if 'presence_penalty' in sig.parameters:
        arguments['presence_penalty'] = request.presence_penalty
    if 'frequency_penalty' in sig.parameters:
        arguments['frequency_penalty'] = request.frequency_penalty
    if 'user' in sig.parameters:
        arguments['user'] = request.user
    if 'best_of' in sig.parameters:
        arguments['best_of'] = request.best_of
    return arguments

async def get_completion_response(
        request: CompletionRequest,
        complete: Complete,
        arguments: dict
    ) -> CompletionResponse:
    async def create_choice(index: int) -> CompletionResponseChoice:
        answer = ''
        try:
            async for chunk in asyncgen(complete, arguments):
                answer += chunk
            return CompletionResponseChoice(
                index=index,
                text=answer,
                finish_reason='stop'
                # TODO: logprobs
            )
        except LengthException as e:
            return CompletionResponseChoice(
                index=index,
                message=answer,
                finish_reason='length'
            )

    choices = [await create_choice(i) for i in range(request.n or 1)] # serial
    prompt_tokens = sum(map(lambda text: tokens(text), request.prompt))
    completion_tokens = sum(map(lambda c: tokens(c.text), choices))

    return CompletionResponse(
        model=request.model,
        choices=choices,
        usage=UsageInfo(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=prompt_tokens + completion_tokens
        )
    )

async def get_completion_stream_response(
        request: CompletionRequest,
        complete: Complete,
        arguments: dict
    ) -> StreamingResponse:
    async def create_stream_choice(index: int):
        try:
            async for chunk in agg(asyncgen(complete, arguments)):
                yield CompletionResponseStreamChoice(
                    index=index,
                    text=chunk,
                    finish_reason=None
                    # TODO: logprobs
                )
            yield CompletionResponseStreamChoice(
                index=index,
                text='',
                finish_reason='stop'
            )
        except LengthException as e:
            yield CompletionResponseStreamChoice(
                index=index,
                text='',
                finish_reason='length'
            )

    all_choices = [
        CompletionResponseStreamChoice(
            index=i,
            text='',
            finish_reason=None
        ) for i in range(request.n or 1)
    ]

    async def stream_response():
        nonlocal all_choices
        for i in range(request.n or 1): # serial
            async for choice in create_stream_choice(i):
                all_choices[i] = choice
                yield 'data: ' + CompletionStreamResponse(
                    model=request.model,
                    choices=all_choices,
                ).model_dump_json() + '\n\n'
        yield 'data: [DONE]\n\n'
    return StreamingResponse(stream_response(), media_type='text/event-stream')

async def create_completions(request: CompletionRequest):
    try:
        complete = g_get_complete()
        if not complete:
            return helpers.create_501_error('complete')

        complete_sig = inspect.signature(complete)
        arguments = get_arguments(request, complete_sig)

        if not request.stream:
            return await get_completion_response(request, complete, arguments)

        return await get_completion_stream_response(request, complete, arguments)
    except Exception as e:
        return helpers.create_500_error(str(e))



from fastapi import APIRouter

from . import chat, complete, embed, protocol, models

def add_routes(router: APIRouter) -> APIRouter:
    router.add_api_route(
        '/models',
        models.get_models,
        methods=['GET'],
        response_model=protocol.ModelList,
        tags=['OpenAI'],
    )

    router.add_api_route(
        '/chat/completions',
        chat.create_chat_completions,
        methods=['POST'],
        response_model=protocol.ChatCompletionResponse,
        tags=['OpenAI'],
    )

    router.add_api_route(
        '/completions',
        complete.create_completions,
        methods=['POST'],
        response_model=protocol.CompletionResponse,
        tags=['OpenAI'],
    )

    router.add_api_route(
        '/embeddings',
        embed.create_embeddings,
        methods=['POST'],
        response_model=protocol.EmbeddingsResponse,
        tags=['OpenAI'],
    )
    return router



from fastapi.testclient import TestClient
import src.svllm as svllm
from tests.models import run_models_tests
from tests.chat import run_chat_tests
from tests.complete import run_complete_tests
from tests.embed import run_embed_tests

svllm.set_chat(svllm.examples.chat)
svllm.set_complete(svllm.examples.complete)
svllm.set_embed(svllm.examples.embed)

client = TestClient(svllm.create_app())

def test_models():
    run_models_tests(client)

def test_chat():
    run_chat_tests(client)

def test_complete():
    run_complete_tests(client)

def test_embed():
    run_embed_tests(client)



#!/usr/bin/env python3

import sys, datetime, argparse, importlib, svllm, uvicorn
from fastapi import Request, Response
from termcolor import colored

def import_from(location: str, name: str):
    module_str, _, attrs_str = location.partition(':')
    if not module_str:
        raise ValueError(f'{name} must be in the format \'module:attr\'.')
    instance = importlib.import_module(module_str)
    for attr in (attrs_str or name).split('.'):
        instance = getattr(instance, attr, None)
    if not instance:
        raise ImportError(f'\'{attrs_str or name}\' not found in \'{module_str}\'')
    return (instance, f'{module_str}:{attrs_str or name}')

def log(message: str, level='INFO'):
    color: str = 'red' if level == 'ERROR' else 'green' if level == 'INFO' else 'yellow'
    time_part = colored(f'[{datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}]', 'cyan')
    print(time_part, f'{colored(level, color)}: {message}')

async def log_request(request: Request):
    log(f'Request {request.method} {request.url}', 'DEBUG')
    try:
        body = await request.body()
        if body:
            log(f'Request body: {body.decode()}', 'DEBUG')
    except Exception as e:
        log(f'Error reading request body: {e}', 'DEBUG')

async def log_response(response: Response):
    log(f'Response {response.status_code}', 'DEBUG')
    if hasattr(response, 'body_iterator'):
        original_body_iterator = response.body_iterator
        async def new_iterator():
            async for chunk in original_body_iterator:
                log(f'Response chunk: {chunk.decode()}', 'DEBUG')
                yield chunk
        response.body_iterator = new_iterator()
    elif hasattr(response, 'body'):
        try:
            log(f'Response body: {response.body.decode()}')
        except Exception as e:
            log(f'Error reading response body: {e}', 'DEBUG')

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--chat', type=str, required=False, help='chat function')
    parser.add_argument('--complete', type=str, required=False, help='complete function')
    parser.add_argument('--embed', type=str, required=False, help='embed function')

    parser.add_argument('--prefix', type=str, default='/v1', required=False, help='api prefix')
    parser.add_argument('--host', type=str, default='127.0.0.1', required=False, help='host address')
    parser.add_argument('--port', type=int, default=5261, required=False, help='port number')
    parser.add_argument('--debug', action='store_true', required=False, help='print debug output')

    cmd_args = parser.parse_args()
    sys.path.insert(0, '.')

    if cmd_args.chat:
        chat, location = import_from(cmd_args.chat, 'chat')
        svllm.base.set_chat(chat)
        log(f'Chat function set to ' + colored(location, 'yellow'))

    if cmd_args.complete:
        complete, location = import_from(cmd_args.complete, 'complete')
        svllm.base.set_complete(complete)
        log(f'Complete function set to ' + colored(location, 'yellow'))

    if cmd_args.embed:
        embed, location = import_from(cmd_args.embed, 'embed')
        svllm.base.set_embed(embed)
        log(f'Embed function set to ' + colored(location, 'yellow'))

    app = svllm.create_app(prefix=cmd_args.prefix)

    if cmd_args.debug:
        @app.middleware('http')
        async def debug_middleware(request: Request, call_next):
            await log_request(request)
            response = await call_next(request)
            await log_response(response)
            return response

    uvicorn.run(app, host=cmd_args.host, port=cmd_args.port)

if __name__ == "__main__":
    main()



import time
from typing import (
    Protocol,
    TypedDict,
    Callable,
    Any,
    Literal,
    Optional,
    List,
    Union,
    Generator,
    AsyncGenerator,
    Coroutine,
    runtime_checkable,
)

class Model(TypedDict):
    id: str
    owned_by: str = 'svllm'
    created: int = int(time.time())

class Message(TypedDict):
    role: Literal['user', 'assistant', 'system']
    content: str

class ModelInfo(Protocol):
    __model__: Optional[str] = None
    __owner__: Optional[str] = None

@runtime_checkable
class ChatSync(ModelInfo, Protocol):
    '''examples:
    def chat(messages: List[Message]) -> str:
        # `callback` can be used to stream response
        # callback('hello world, ')
        return 'hello world again'
    '''
    def __call__(
        self,
        messages: List[Message],
        model: str,
        temperature: Optional[float] = 0.7,
        top_p: Optional[float] = 1.0,
        top_k: Optional[int] = -1,
        max_tokens: Optional[int] = None,
        stop: Optional[Union[str, List[str]]] = None,
        stream: Optional[bool] = False,
        presence_penalty: Optional[float] = 0.0,
        frequency_penalty: Optional[float] = 0.0,
        user: Optional[str] = None,
        callback: Optional[Callable[[str], Any]] = None,
    ) -> str:
        ...

@runtime_checkable
class ChatGenerator(ModelInfo, Protocol):
    '''examples:
    def chat(messages: List[Message]) -> Generator[str, Any, Optional[str]]:
        for chunk in ['hello', ' ', 'world']:
            yield chunk
    '''
    def __call__(
        self,
        messages: List[Message],
        model: str,
        temperature: Optional[float] = 0.7,
        top_p: Optional[float] = 1.0,
        top_k: Optional[int] = -1,
        max_tokens: Optional[int] = None,
        stop: Optional[Union[str, List[str]]] = None,
        stream: Optional[bool] = False,
        presence_penalty: Optional[float] = 0.0,
        frequency_penalty: Optional[float] = 0.0,
        user: Optional[str] = None,
    ) -> Generator[str, Any, Optional[str]]:
        ...

@runtime_checkable
class ChatCoroutine(ModelInfo, Protocol):
    '''examples:
    async def chat(messages: List[Message]) -> Coroutine[Any, Any, str]:
        return 'hello world'
    '''
    async def __call__(
        self,
        messages: List[Message],
        model: str,
        temperature: Optional[float] = 0.7,
        top_p: Optional[float] = 1.0,
        top_k: Optional[int] = -1,
        max_tokens: Optional[int] = None,
        stop: Optional[Union[str, List[str]]] = None,
        stream: Optional[bool] = False,
        presence_penalty: Optional[float] = 0.0,
        frequency_penalty: Optional[float] = 0.0,
        user: Optional[str] = None,
        callback: Optional[Callable[[str], Any]] = None,
    ) -> Coroutine[Any, Any, str]:
        ...

@runtime_checkable
class ChatAsyncGenerator(ModelInfo, Protocol):
    '''examples:
    async def chat(messages: List[Message]) -> AsyncGenerator[str, str]:
        for chunk in ['hello', ' ', 'world']:
            yield chunk
    '''
    async def __call__(
        self,
        messages: List[Message],
        model: str,
        temperature: Optional[float] = 0.7,
        top_p: Optional[float] = 1.0,
        top_k: Optional[int] = -1,
        max_tokens: Optional[int] = None,
        stop: Optional[Union[str, List[str]]] = None,
        stream: Optional[bool] = False,
        presence_penalty: Optional[float] = 0.0,
        frequency_penalty: Optional[float] = 0.0,
        user: Optional[str] = None,
    ) -> AsyncGenerator[str, str]:
        ...

Chat = Union[
    ChatSync,
    ChatGenerator,
    ChatCoroutine,
    ChatAsyncGenerator,
]

@runtime_checkable
class CompleteSync(ModelInfo, Protocol):
    '''examples:
    def complete(prompt: str) -> str:
        # `callback` can be used to stream response
        # callback('hello world, ')
        return 'hello world again'
    '''
    def __call__(
        self,
        prompt: str,
        model: str,
        suffix: Optional[str] = None,
        temperature: Optional[float] = 0.7,
        top_p: Optional[float] = 1.0,
        top_k: Optional[int] = -1,
        max_tokens: Optional[int] = None,
        stop: Optional[Union[str, List[str]]] = None,
        stream: Optional[bool] = False,
        logprobs: Optional[int] = None,
        echo: Optional[bool] = False,
        presence_penalty: Optional[float] = 0.0,
        frequency_penalty: Optional[float] = 0.0,
        user: Optional[str] = None,
        best_of: Optional[int] = None,
        callback: Optional[Callable[[str], Any]] = None,
    ) -> str:
        ...

@runtime_checkable
class CompleteGenerator(ModelInfo, Protocol):
    '''examples:
    def complete(prompt: str) -> Generator[str, Any, Optional[str]]:
        for chunk in ['hello', ' ', 'world']:
            yield chunk
    '''
    def __call__(
        self,
        prompt: str,
        model: str,
        suffix: Optional[str] = None,
        temperature: Optional[float] = 0.7,
        top_p: Optional[float] = 1.0,
        top_k: Optional[int] = -1,
        max_tokens: Optional[int] = None,
        stop: Optional[Union[str, List[str]]] = None,
        stream: Optional[bool] = False,
        logprobs: Optional[int] = None,
        echo: Optional[bool] = False,
        presence_penalty: Optional[float] = 0.0,
        frequency_penalty: Optional[float] = 0.0,
        user: Optional[str] = None,
        best_of: Optional[int] = None,
    ) -> Generator[str, Any, Optional[str]]:
        ...

@runtime_checkable
class CompleteCoroutine(ModelInfo, Protocol):
    '''examples:
    async def complete(prompt: str) -> Coroutine[Any, Any, str]:
        return 'hello world'
    '''
    async def __call__(
        self,
        prompt: str,
        model: str,
        suffix: Optional[str] = None,
        temperature: Optional[float] = 0.7,
        top_p: Optional[float] = 1.0,
        top_k: Optional[int] = -1,
        max_tokens: Optional[int] = None,
        stop: Optional[Union[str, List[str]]] = None,
        stream: Optional[bool] = False,
        logprobs: Optional[int] = None,
        echo: Optional[bool] = False,
        presence_penalty: Optional[float] = 0.0,
        frequency_penalty: Optional[float] = 0.0,
        user: Optional[str] = None,
        best_of: Optional[int] = None,
        callback: Optional[Callable[[str], Any]] = None,
    ) -> Coroutine[Any, Any, str]:
        ...

@runtime_checkable
class CompleteAsyncGenerator(ModelInfo, Protocol):
    '''examples:
    async def complete(prompt: str) -> AsyncGenerator[str, str]:
        for chunk in ['hello', ' ', 'world']:
            yield chunk
    '''
    async def __call__(
        self,
        prompt: str,
        model: str,
        suffix: Optional[str] = None,
        temperature: Optional[float] = 0.7,
        top_p: Optional[float] = 1.0,
        top_k: Optional[int] = -1,
        max_tokens: Optional[int] = None,
        stop: Optional[Union[str, List[str]]] = None,
        stream: Optional[bool] = False,
        logprobs: Optional[int] = None,
        echo: Optional[bool] = False,
        presence_penalty: Optional[float] = 0.0,
        frequency_penalty: Optional[float] = 0.0,
        user: Optional[str] = None,
        best_of: Optional[int] = None,
    ) -> AsyncGenerator[str, str]:
        ...

Complete = Union[
    CompleteSync,
    CompleteGenerator,
    CompleteCoroutine,
    CompleteAsyncGenerator,
]

@runtime_checkable
class EmbedSync(ModelInfo, Protocol):
    '''examples:
    def embed(input: List[str]) -> List[List[float]]:
        return [[0.1, 0.2, 0.3], ...]
    '''
    def __call__(
        self,
        input: List[str],
        model: str,
        engine: Optional[str] = None,
        user: Optional[str] = None,
        encoding_format: Optional[str] = None,
    ) -> List[List[float]]:
        ...

@runtime_checkable
class EmbedAsync(ModelInfo, Protocol):
    '''examples:
    async def embed(input: List[str]) -> Coroutine[Any, Any, List[List[float]]]:
        return [[0.1, 0.2, 0.3], ...]
    '''
    async def __call__(
        self,
        input: List[str],
        model: str,
        engine: Optional[str] = None,
        user: Optional[str] = None,
        encoding_format: Optional[str] = None,
    ) -> List[List[float]]:
        ...

Embed = Union[
    EmbedSync,
    EmbedAsync,
]

# Exception for token length exceeding
class LengthException(Exception):
    def __init__(self, message: str = ''):
        super().__init__(message)
        self.message = message



#!/usr/bin/env python3

import argparse, os, sys, json, httpx
from termcolor import colored

def get_request(base_url: str, key: str):
    headers = { 'Authorization': f'Bearer {key}' } if key else {}
    return httpx.Client(base_url=base_url, headers=headers, timeout=30)

def get_input():
    content = ''
    while not content:
        try:
            content = input(colored('>>> ', 'green'))
        except EOFError:
            print(colored('switch to multiline mode, end with Ctrl+D/EOF', 'green'))
            print(colored('>>>> ', 'cyan'), end='', flush=True)
            content = sys.stdin.read()
            print()
            break
    return content

chat_history: list[dict] = []

def run_chat(request: httpx.Client, model: str, stream = True, history = True):
    url = f'/chat/completions'

    messages = [{ 'role': 'user', 'content': get_input() }]
    if history:
        chat_history.extend(messages)
        messages = chat_history
    payload = { 'model': model, 'messages': messages, 'stream': stream }

    if not stream:
        content_text = ''
        response = request.post(url, json=payload)
        if response.status_code != 200:
            if history:
                chat_history.pop()
            print(colored(f'{json.dumps(response.json(), indent=2)}', 'red'))
            return

        data = response.json()
        if 'choices' in data and len(data['choices']) > 0:
            choice = data['choices'][0]
            if 'message' in choice and 'content' in choice['message']:
                content_text = choice['message']['content']
        print(content_text)
        if history:
            chat_history.append({ 'role': 'assistant', 'content': content_text })
        return

    with request.stream(method='POST', url=url, json=payload) as response:
        if response.status_code != 200:
            if history:
                chat_history.pop()
            body = ''.join([chunk for chunk in response.iter_lines()])
            print(colored(f'{json.dumps(json.loads(body), indent=2)}', 'red'))
            return

        content_text = ''
        for chunk in response.iter_lines():
            chunk = chunk.strip()
            if chunk.startswith('data:'):
                chunk = chunk[5:].strip()
            if not chunk or chunk == '[DONE]':
                continue
            data = json.loads(chunk)
            if not data.get('choices'):
                continue
            choice = data['choices'][0]
            if 'delta' in choice and 'content' in choice['delta']:
                content_text += choice['delta']['content']
                print(choice['delta']['content'], end='', flush=True)
        print()
        if history:
            chat_history.append({ 'role': 'assistant', 'content': content_text })

def run_completion(request: httpx.Client, model: str, stream = True):
    url = f'/completions'
    payload = { 'model': model, 'prompt': get_input(), 'stream': stream }

    if not stream:
        content_text = ''
        response = request.post(url, json=payload)
        if response.status_code != 200:
            print(colored(f'{json.dumps(response.json(), indent=2)}', 'red'))
            return

        data = response.json()
        if 'choices' in data and len(data['choices']) > 0:
            choice = data['choices'][0]
            if 'text' in choice:
                content_text = choice['text']
        print(content_text)
        return

    with request.stream(method='POST', url=url, json=payload) as response:
        if response.status_code != 200:
            body = ''.join([chunk for chunk in response.iter_lines()])
            print(colored(f'{json.dumps(json.loads(body), indent=2)}', 'red'))
            return
        for chunk in response.iter_lines():
            chunk = chunk.strip()
            if chunk.startswith('data:'):
                chunk = chunk[5:].strip()
            if not chunk or chunk == '[DONE]':
                continue
            data = json.loads(chunk)
            if not data.get('choices'):
                continue
            choice = data['choices'][0]
            if 'text' in choice:
                print(choice['text'], end='', flush=True)
        print()


def run_embed(request: httpx.Client, model: str):
    url = f'/embeddings'
    payload = { 'model': model, 'input': get_input() }

    response = request.post(url, json=payload)
    if response.status_code != 200:
        print(colored(f'{json.dumps(response.json(), indent=2)}', 'red'))
        return

    data = response.json()
    if 'data' in data and len(data['data']) > 0:
        embedding = data['data'][0].get('embedding')
        if embedding:
            print(embedding)

def main():
    parser = argparse.ArgumentParser(epilog='Tips: you can switch to multiline mode using Ctrl+D/EOF.')
    parser.add_argument('--model', type=str, required=False, help='model name')
    parser.add_argument('--chat', action='store_true', help='chat mode (default)')
    parser.add_argument('--complete', action='store_true', help='completion mode')
    parser.add_argument('--embed', action='store_true', help='embedding mode')
    parser.add_argument('--no-stream', action='store_true', help='no stream mode (chat/complete mode)')
    parser.add_argument('--no-history', action='store_true', help='no history (chat mode)')

    api_key = os.environ.get('SVLLM_API_KEY')
    parser.add_argument('--key', type=str, required=False, default=api_key, help='API key for authentication')
    parser.add_argument('base_url', type=str, nargs='?', default='http://127.0.0.1:5261/v1', help='base URL for the API')

    cmd_args = parser.parse_args()

    if not cmd_args.chat and not cmd_args.complete and not cmd_args.embed:
        cmd_args.chat = True
    if not cmd_args.model:
        cmd_args.model = 'svllm-chat' if cmd_args.chat else \
                         'svllm-complete' if cmd_args.complete else \
                         'svllm-embed'

    request = get_request(cmd_args.base_url, cmd_args.key)
    if cmd_args.chat:
        print(colored(f'Chat mode with model: {cmd_args.model}', 'blue'))
        while True:
            try:
                run_chat(request, cmd_args.model, not cmd_args.no_stream, not cmd_args.no_history)
            except KeyboardInterrupt:
                print(colored('\nExiting chat mode.', 'red'))
                break
    elif cmd_args.complete:
        print(colored(f'Completion mode with model: {cmd_args.model}', 'blue'))
        while True:
            try:
                run_completion(request, cmd_args.model, not cmd_args.no_stream)
            except KeyboardInterrupt:
                print(colored('\nExiting completion mode.', 'red'))
                break
    elif cmd_args.embed:
        print(colored(f'Embedding mode with model: {cmd_args.model}', 'blue'))
        while True:
            try:
                run_embed(request, cmd_args.model)
            except KeyboardInterrupt:
                print(colored('\nExiting embedding mode.', 'red'))
                break

if __name__ == "__main__":
    main()



import inspect
from typing import List
from fastapi.responses import StreamingResponse

from ..protocol import LengthException, Message, Chat
from ..base import get_chat as g_get_chat
from ..helpers import agg, asyncgen, tokens
from . import helpers
from .protocol import (
    ChatCompletionRequest,
    ChatCompletionResponse,
    ChatCompletionResponseChoice,
    ChatCompletionResponseStreamChoice,
    ChatCompletionStreamResponse,
    ChatMessage,
    DeltaMessage,
    UsageInfo,
)

def get_messages(request: ChatCompletionRequest) -> List[Message]:
    openai_messages = request.messages
    if isinstance(openai_messages, str):
        return [{ 'role': 'user', 'content': openai_messages }]
    elif isinstance(openai_messages, list):
        messages: List[Message] = []
        for (idx, msg) in enumerate(openai_messages):
            if 'content' not in msg:
                raise ValueError(f'Message at index {idx} is missing \'content\' field.')
            # ignore no `text` type` content now
            if isinstance(msg['content'], dict) and msg['content'].get('type') != 'text':
                continue
            if isinstance(msg['content'], dict) and 'text' not in msg['content']:
                raise ValueError(f'Message at index {idx} is missing \'text\' field in \'content\'.')

            # we need also handle `system` roles correctly here
            prev_role = messages[idx - 1]['role'] if idx > 0 else None
            role = msg.get('role', 'user' if prev_role != 'user' else 'assistant')
            content = msg['content'] if isinstance(msg['content'], str) else msg['content']['text']
            messages.append({ 'role': role, 'content': content })
        return messages

def get_arguments(request: ChatCompletionRequest, sig: inspect.Signature) -> dict:
    arguments = {}
    if 'model' in sig.parameters:
        arguments['model'] = request.model
    if 'temperature' in sig.parameters:
        arguments['temperature'] = request.temperature
    if 'top_p' in sig.parameters:
        arguments['top_p'] = request.top_p
    if 'top_k' in sig.parameters:
        arguments['top_k'] = request.top_k
    if 'max_tokens' in sig.parameters:
        arguments['max_tokens'] = request.max_tokens
    if 'stop' in sig.parameters:
        arguments['stop'] = request.stop
    if 'stream' in sig.parameters:
        arguments['stream'] = request.stream
    if 'presence_penalty' in sig.parameters:
        arguments['presence_penalty'] = request.presence_penalty
    if 'frequency_penalty' in sig.parameters:
        arguments['frequency_penalty'] = request.frequency_penalty
    if 'user' in sig.parameters:
        arguments['user'] = request.user
    return arguments

async def get_chat_completion_response(
        request: ChatCompletionRequest,
        chat: Chat,
        arguments: dict,
        messages: List[Message] = []
    ) -> ChatCompletionResponse:
    async def create_choice(index: int) -> ChatCompletionResponseChoice:
        answer = ''
        try:
            async for chunk in asyncgen(chat, arguments):
                answer += chunk
            return ChatCompletionResponseChoice(
                index=index,
                message=ChatMessage(role='assistant', content=answer),
                finish_reason='stop'
            )
        except LengthException as e:
            return ChatCompletionResponseChoice(
                index=index,
                message=ChatMessage(role='assistant', content=answer),
                finish_reason='length'
            )

    choices = [await create_choice(i) for i in range(request.n or 1)] # serial
    prompt_tokens = sum(map(lambda m: tokens(m.get('content', '')), messages))
    completion_tokens = sum(map(lambda c: tokens(c.message.content), choices))

    return ChatCompletionResponse(
        model=request.model,
        choices=choices,
        usage=UsageInfo(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=prompt_tokens + completion_tokens
        )
    )

async def get_chat_completion_stream_response(
        request: ChatCompletionRequest,
        chat: Chat,
        arguments: dict
    ) -> StreamingResponse:
    async def create_stream_choice(index: int):
        try:
            async for chunk in agg(asyncgen(chat, arguments)):
                yield ChatCompletionResponseStreamChoice(
                    index=index,
                    delta=DeltaMessage(role='assistant', content=chunk),
                    finish_reason=None
                )
            yield ChatCompletionResponseStreamChoice(
                index=index,
                delta=DeltaMessage(),
                finish_reason='stop'
            )
        except LengthException as e:
            yield ChatCompletionResponseStreamChoice(
                index=index,
                delta=DeltaMessage(),
                finish_reason='length'
            )

    all_choices = [
        ChatCompletionResponseStreamChoice(
            index=i,
            delta=DeltaMessage(),
            finish_reason=None
        ) for i in range(request.n or 1)
    ]

    async def stream_response():
        nonlocal all_choices
        for i in range(request.n or 1): # serial
            async for choice in create_stream_choice(i):
                all_choices[i] = choice
                yield 'data: ' + ChatCompletionStreamResponse(
                    model=request.model,
                    choices=all_choices,
                ).model_dump_json() + '\n\n'
        yield 'data: [DONE]\n\n'
    return StreamingResponse(stream_response(), media_type='text/event-stream')

async def create_chat_completions(request: ChatCompletionRequest):
    try:
        chat = g_get_chat()
        if not chat:
            return helpers.create_501_error('chat')

        chat_sig = inspect.signature(chat)
        arguments = get_arguments(request, chat_sig)

        messages = get_messages(request)
        if 'messages' in chat_sig.parameters:
            arguments['messages'] = messages

        if not request.stream:
            return await get_chat_completion_response(request, chat, arguments, messages)

        return await get_chat_completion_stream_response(request, chat, arguments)
    except ValueError as e:
        return helpers.create_500_error(str(e))



from fastapi.testclient import TestClient

def run_models_tests(client: TestClient):
    print('Running models tests...')
    response = client.get('/v1/models')
    assert response.status_code == 200
    assert 'data' in response.json()
    assert isinstance(response.json()['data'], list)
    assert len(response.json()['data']) == 3
    assert response.json()['data'][0]['id'] == 'svllm-chat'
    assert response.json()['data'][1]['id'] == 'svllm-complete'
    assert response.json()['data'][2]['id'] == 'svllm-embed'
